#include "Generator.hpp"
#include "Program.hpp"
#include "Problem.hpp"
#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <vector>

using std::vector;

bool generateImpl(Program& program,
                  uint32_t len,
                  uint32_t usedOps,
                  uint32_t usedOpsCount,
                  uint32_t stackSize,
                  bool folding,
                  bool tfolding,
                  bool needFold,
                  const GenCallback& callback)
{
    Op* ops = program.m_ops;
    uint32_t iOp = program.m_foldOpEnd;

    if ((usedOpsCount + (needFold ? 1 : 0)) > (len - iOp))
    {
        return true;
    }

    if (usedOps & UsedOp::If0)
    {
        if (stackSize > 1 + 2 * (len - iOp))
        {
            return true;
        }
    }
    else if (usedOps & (UsedOp::And |
                        UsedOp::Or |
                        UsedOp::Xor |
                        UsedOp::Plus |
                        UsedOp::TFold |
                        UsedOp::Fold))
    {
        if (stackSize > 1 + (len - iOp))
        {
            return true;
        }
    }
    else
    {
        if (stackSize > 1)
        {
            return true;
        }
    }

    if (iOp == len)
    {
        if (stackSize == 1 && !needFold)
        {
            return callback(program);
        }
        else
        {
            printf("Error!\n");
            return true;
        }
    }

    if (stackSize == 1 && needFold)
    {
        if (!generateImpl(program, len, usedOps, usedOpsCount, 0, true, false, false, callback))
            return false;
    }

    if (!folding)
    {
        program.m_opEnd++;
    }
    program.m_foldOpEnd++;

    {
        ops[iOp] = Op::Zero;
        if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize + 1, folding, tfolding, needFold, callback))
            return false;

        ops[iOp] = Op::One;
        if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize + 1, folding, tfolding, needFold, callback))
            return false;

        if (!tfolding)
        {
            ops[iOp] = Op::X;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize + 1, folding, tfolding, needFold, callback))
                return false;
        }

        if (folding)
        {
            ops[iOp] = Op::Y;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize + 1, folding, tfolding, needFold, callback))
                return false;

            ops[iOp] = Op::Z;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize + 1, folding, tfolding, needFold, callback))
                return false;
        }
    }

    if (stackSize >= 1)
    {
        if (usedOps & UsedOp::Not)
        {
            ops[iOp] = Op::Not;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize, folding, tfolding, needFold, callback))
                return false;
            if (!generateImpl(program, len, usedOps & ~UsedOp::Not, usedOpsCount - 1, stackSize, folding, tfolding, needFold, callback))
                return false;
        }

        if (usedOps & UsedOp::Shl1)
        {
            ops[iOp] = Op::Shl1;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize, folding, tfolding, needFold, callback))
                return false;
            if (!generateImpl(program, len, usedOps & ~UsedOp::Shl1, usedOpsCount - 1, stackSize, folding, tfolding, needFold, callback))
                return false;
        }

        if (usedOps & UsedOp::Shr1)
        {
            ops[iOp] = Op::Shr1;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize, folding, tfolding, needFold, callback))
                return false;
            if (!generateImpl(program, len, usedOps & ~UsedOp::Shr1, usedOpsCount - 1, stackSize, folding, tfolding, needFold, callback))
                return false;
        }

        if (usedOps & UsedOp::Shr4)
        {
            ops[iOp] = Op::Shr4;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize, folding, tfolding, needFold, callback))
                return false;
            if (!generateImpl(program, len, usedOps & ~UsedOp::Shr4, usedOpsCount - 1, stackSize, folding, tfolding, needFold, callback))
                return false;
        }

        if (usedOps & UsedOp::Shr16)
        {
            ops[iOp] = Op::Shr16;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize, folding, tfolding, needFold, callback))
                return false;
            if (!generateImpl(program, len, usedOps & ~UsedOp::Shr16, usedOpsCount - 1, stackSize, folding, tfolding, needFold, callback))
                return false;
        }
    }

    if (stackSize >= 2)
    {
        if (usedOps & UsedOp::And)
        {
            ops[iOp] = Op::And;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize - 1, folding, tfolding, needFold, callback))
                return false;
            if (!generateImpl(program, len, usedOps & ~UsedOp::And, usedOpsCount - 1, stackSize - 1, folding, tfolding, needFold, callback))
                return false;
        }

        if (usedOps & UsedOp::Or)
        {
            ops[iOp] = Op::Or;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize - 1, folding, tfolding, needFold, callback))
                return false;
            if (!generateImpl(program, len, usedOps & ~UsedOp::Or, usedOpsCount - 1, stackSize - 1, folding, tfolding, needFold, callback))
                return false;
        }

        if (usedOps & UsedOp::Xor)
        {
            ops[iOp] = Op::Xor;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize - 1, folding, tfolding, needFold, callback))
                return false;
            if (!generateImpl(program, len, usedOps & ~UsedOp::Xor, usedOpsCount - 1, stackSize - 1, folding, tfolding, needFold, callback))
                return false;
        }

        if (usedOps & UsedOp::Plus)
        {
            ops[iOp] = Op::Plus;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize - 1, folding, tfolding, needFold, callback))
                return false;
            if (!generateImpl(program, len, usedOps & ~UsedOp::Plus, usedOpsCount - 1, stackSize - 1, folding, tfolding, needFold, callback))
                return false;
        }
    }

    if (stackSize == 2)
    {
        if (usedOps & UsedOp::Fold)
        {
            ops[iOp] = Op::Fold;
            if (!generateImpl(program, len, usedOps & ~UsedOp::Fold, usedOpsCount - 1, stackSize - 1, folding, tfolding, true, callback))
                return false;
        }
    }

    if (stackSize >= 3)
    {
        if (usedOps & UsedOp::If0)
        {
            ops[iOp] = Op::If0;
            if (!generateImpl(program, len, usedOps, usedOpsCount, stackSize - 2, folding, tfolding, needFold, callback))
                return false;
            if (!generateImpl(program, len, usedOps & ~UsedOp::If0, usedOpsCount - 1, stackSize - 2, folding, tfolding, needFold, callback))
                return false;
        }
    }

    if (!folding)
    {
        program.m_opEnd--;
    }
    program.m_foldOpEnd--;

    return true;
}

void generate(uint32_t size, uint32_t usedOps, const GenCallback& callback)
{
    Program program;

    uint32_t specialCount = 0;
    if (usedOps & UsedOp::Fold) specialCount++;
    if (usedOps & UsedOp::TFold) specialCount++;
    if (usedOps & UsedOp::Bonus) specialCount++;
    if (specialCount > 1)
    {
        printf("Invalid special operator combination\n");
        return;
    }

    uint32_t usedOpsCount = 0;
    if (usedOps & UsedOp::Not) usedOpsCount++;
    if (usedOps & UsedOp::Shl1) usedOpsCount++;
    if (usedOps & UsedOp::Shr1) usedOpsCount++;
    if (usedOps & UsedOp::Shr4) usedOpsCount++;
    if (usedOps & UsedOp::Shr16) usedOpsCount++;
    if (usedOps & UsedOp::And) usedOpsCount++;
    if (usedOps & UsedOp::Or) usedOpsCount++;
    if (usedOps & UsedOp::Xor) usedOpsCount++;
    if (usedOps & UsedOp::Plus) usedOpsCount++;
    if (usedOps & UsedOp::If0) usedOpsCount++;
    //if (usedOps & UsedOp::TFold) usedOpsCount++;
    if (usedOps & UsedOp::Fold) usedOpsCount++;

    if (usedOps & UsedOp::TFold)
    {
        program.m_ops[0] = Op::X;
        program.m_ops[1] = Op::Zero;
        program.m_ops[2] = Op::Fold;
        program.m_opEnd = 3;
        program.m_foldOpEnd = 3;

        generateImpl(program, size - 2, usedOps & ~(UsedOp::TFold), usedOpsCount, 0, true, true, false, callback);
        return;
    }

    if (usedOps & UsedOp::Fold)
    {
        generateImpl(program, size - 2, usedOps, usedOpsCount, 0, false, false, false, callback);
        return;
    }

    generateImpl(program, size - 1, usedOps & ~(UsedOp::Bonus), usedOpsCount, 0, false, false, false, callback);
}

class Range
{
public:
    Range(count_t low = 0, count_t high = ~(count_t)0)
        : m_low(low), m_high(high) {}

    Range& operator -=(count_t delta) {
        if (delta >= m_low)
        {
            m_low = 0;
        }
        else
        {
            m_low -= delta;
        }

        if (delta >= m_high)
        {
            m_high = 0;
        }
        else
        {
            m_high -= delta;
        }

        return *this;
    }

    Range& operator +=(count_t delta) {
        if (delta >= (~(count_t)0) - m_low)
        {
            m_low = ~(count_t)0;
        }
        else
        {
            m_low += delta;
        }

        if (delta >= (~(count_t)0) - m_high)
        {
            m_high = ~(count_t)0;
        }
        else
        {
            m_high += delta;
        }

        return *this;
    }

    bool active(count_t count) const {
        return (0 < m_high) && (count > m_low) && (m_high > m_low) && (count > 0);
    }

    count_t m_low;
    count_t m_high;
};

enum class Mode : uint32_t
{
    Simple,
    TBody,
    Body,
    Fold,

    Count
};
const uint32_t ModeCount = (uint32_t)Mode::Count;

enum class Context : uint32_t
{
    Top,
    Not,
    Shl1,
    Shr1,
    Shr4,
    Shr16,
    Shr1Shl1,
    Shr4Shl1,
    Shr16Shl1,
    And,
    Or,
    Xor,
    AndNot,
    OrNot,
    Plus,
    If,
    ThenElse,
    FoldBytes,
    FoldAccum,
    FoldBody,

    Count,

    First = Top,
    Last = Count - 1
};
const uint32_t ContextCount = (uint32_t)Context::Count;

class GenImpl
{
public:
    GenImpl();
    void init(uint32_t sizeIn, uint32_t usedOpsIn);
    count_t getCount();
    bool generate(const GenCallback& callback);

private:
    void genCounts();
    count_t getCountFoldBody(uint32_t size);
    bool genFoldBody(uint32_t size, Range range, const GenCallback& callback);
    bool gen(uint32_t size, Context context, Mode mode, Range range, const GenCallback& callback);

    bool bAllSizes;
    bool bAllOps;
    bool bEasyOp2;
    bool bEasyIf0;
    bool bFoldOnlyYZ;
    bool bGenFoldBody;
    bool bFoldFirst;

    Program program;
    uint32_t sizeIn;
    uint32_t usedOpsIn;
    uint32_t usedOps;
    uint32_t pos;
    bool bSimple;
    bool bTFold;
    bool bFold;
    bool bBonus;
    count_t counts[50][ContextCount][ModeCount];
};

GenImpl::GenImpl()
    : sizeIn(0),
      usedOpsIn(0),
      usedOps(0),
      pos(0),
      bSimple(false),
      bTFold(false),
      bFold(false),
      bBonus(false)
{
}

void GenImpl::init(uint32_t sizeIn_, uint32_t usedOpsIn_)
{
    bAllSizes = true;
    bAllOps = false;
    bEasyOp2 = false;
    bEasyIf0 = false;
    bFoldOnlyYZ = true;
    bGenFoldBody = true;
    bFoldFirst = true;

    sizeIn = sizeIn_;
    usedOpsIn = usedOpsIn_;
    pos = 0;

    usedOps = usedOpsIn & ~(UsedOp::TFold | UsedOp::Fold | UsedOp::Bonus);
    if (bAllOps)
    {
        usedOps = 
            UsedOp::Not |
            UsedOp::Shl1 |
            UsedOp::Shr1 |
            UsedOp::Shr4 |
            UsedOp::Shr16 |
            UsedOp::And |
            UsedOp::Or |
            UsedOp::Xor |
            UsedOp::Plus |
            UsedOp::If0;
    }

    bSimple = false;
    bTFold = false;
    bFold = false;
    bBonus = false;
    if (usedOpsIn & UsedOp::TFold)
    {
        bTFold = true;
    }
    else if (usedOpsIn & UsedOp::Fold)
    {
        bFold = true;
    }
    else if (usedOpsIn & UsedOp::Bonus)
    {
        bBonus = true;
    }
    else
    {
        bSimple = true;
    }

    genCounts();
}

void GenImpl::genCounts()
{
    uint32_t maxSize = sizeIn - 1;

    for (uint32_t size = 0; size <= maxSize; size++)
    for (uint32_t iContext = (uint32_t)Context::First; iContext <= (uint32_t)Context::Last; iContext++)
    {
        Context context = (Context)iContext;

        count_t& simpleCount = counts[size][iContext][(uint32_t)Mode::Simple];
        count_t& tbodyCount = counts[size][iContext][(uint32_t)Mode::TBody];
        count_t& bodyCount = counts[size][iContext][(uint32_t)Mode::Body];
        count_t& foldCount = counts[size][iContext][(uint32_t)Mode::Fold];

        simpleCount = 0;
        tbodyCount = 0;
        bodyCount = 0;
        foldCount = 0;

        if (size == 1 &&
            (!(usedOpsIn & UsedOp::Bonus) || context != Context::Top))
        {
            // Zero
            if (context != Context::Shl1 &&
                context != Context::Shr1 &&
                context != Context::Shr4 &&
                context != Context::Shr16 &&
                context != Context::Shr1Shl1 &&
                context != Context::Shr4Shl1 &&
                context != Context::Shr16Shl1 &&
                context != Context::And &&
                context != Context::Or &&
                context != Context::Xor &&
                context != Context::AndNot &&
                context != Context::OrNot &&
                context != Context::Plus &&
                context != Context::If)
            {
                simpleCount++;
                tbodyCount++;
                bodyCount++;
            }

            // One
            if (context != Context::Shr1 &&
                context != Context::Shr4 &&
                context != Context::Shr16 &&
                context != Context::Shr1Shl1 &&
                context != Context::Shr4Shl1 &&
                context != Context::Shr16Shl1 &&
                context != Context::If)
            {
                simpleCount++;
                tbodyCount++;
                bodyCount++;
            }

            // X
            simpleCount++;
            bodyCount++;

            // Y
            tbodyCount++;
            bodyCount++;

            // Z
            tbodyCount++;
            bodyCount++;
        }

        // Op1
        if (size >= 2 &&
            (!(usedOpsIn & UsedOp::Bonus) || context != Context::Top))
        {
            auto doOp1 = [&](uint32_t usedOp, Context context) {
                if (!(usedOps & usedOp)) return;

                uint32_t argSize = size - 1;
                simpleCount += counts[argSize][(uint32_t)context][(uint32_t)Mode::Simple];
                tbodyCount += counts[argSize][(uint32_t)context][(uint32_t)Mode::TBody];
                bodyCount += counts[argSize][(uint32_t)context][(uint32_t)Mode::Body];
                if (!bFoldFirst)
                    foldCount += counts[argSize][(uint32_t)context][(uint32_t)Mode::Fold];
            };

            // Not
            if (context != Context::Not &&
                context != Context::Xor &&
                context != Context::AndNot &&
                context != Context::OrNot &&
                context != Context::If)
            {
                Context newContext = Context::Not;
                switch (context)
                {
                case Context::And: newContext = Context::AndNot; break;
                case Context::Or: newContext = Context::OrNot; break;
                default: ;
                }
                doOp1(UsedOp::Not, newContext);
            }

            // Shl1
            {
                Context newContext = Context::Shl1;
                switch (context)
                {
                case Context::Shr1: newContext = Context::Shr1Shl1; break;
                case Context::Shr4: newContext = Context::Shr4Shl1; break;
                case Context::Shr16: newContext = Context::Shr16Shl1; break;
                default: ;
                }
                doOp1(UsedOp::Shl1, newContext);
            }

            // Shr1
            if (context != Context::Shr16 &&
                context != Context::Shr4 &&
                context != Context::Shr1Shl1 &&
                context != Context::Shr4Shl1 &&
                context != Context::Shr16Shl1)
            {
                doOp1(UsedOp::Shr1, Context::Shr1);
            }

            // Shr4
            if (context != Context::Shr16)
            {
                doOp1(UsedOp::Shr4, Context::Shr4);
            }

            // Shr16
            doOp1(UsedOp::Shr16, Context::Shr16);
        }

        // Op2
        if (size >= 3 &&
            (!(usedOpsIn & UsedOp::Bonus) || context != Context::Top))
        {
            auto doOp2 = [&](uint32_t usedOp, Context context, bool allowEqual) {
                uint32_t iContext = (uint32_t)context;
                if (!(usedOps & usedOp)) return;

                for (uint32_t leftSize = 1; leftSize + 1 < (bEasyOp2 ? 3 : size); leftSize++)
                {
                    uint32_t rightSize = size - 1 - leftSize;

                    if (leftSize > rightSize)
                    {
                        break;
                    }
                    else if (leftSize == rightSize)
                    {
                        count_t simpleArgCount = counts[leftSize][iContext][(uint32_t)Mode::Simple];
                        count_t tbodyArgCount = counts[leftSize][iContext][(uint32_t)Mode::TBody];
                        count_t bodyArgCount = counts[leftSize][iContext][(uint32_t)Mode::Body];
                        if (!allowEqual)
                        {
                            simpleArgCount--;
                            tbodyArgCount--;
                            bodyArgCount--;
                        }

                        simpleCount += (simpleArgCount * (simpleArgCount + 1)) / 2;
                        tbodyCount += (tbodyArgCount * (tbodyArgCount + 1)) / 2;
                        bodyCount += (bodyArgCount * (bodyArgCount + 1)) / 2;
                        if (!bFoldFirst)
                            foldCount +=
                                counts[leftSize][iContext][(uint32_t)Mode::Simple] *
                                counts[rightSize][iContext][(uint32_t)Mode::Fold];
                    }
                    else
                    {
                        simpleCount +=
                            counts[leftSize][iContext][(uint32_t)Mode::Simple] *
                            counts[rightSize][iContext][(uint32_t)Mode::Simple];

                        tbodyCount +=
                            counts[leftSize][iContext][(uint32_t)Mode::TBody] *
                            counts[rightSize][iContext][(uint32_t)Mode::TBody];

                        bodyCount +=
                            counts[leftSize][iContext][(uint32_t)Mode::Body] *
                            counts[rightSize][iContext][(uint32_t)Mode::Body];

                        if (!bFoldFirst)
                            foldCount +=
                                counts[leftSize][iContext][(uint32_t)Mode::Fold] *
                                counts[rightSize][iContext][(uint32_t)Mode::Simple] +
                                counts[leftSize][iContext][(uint32_t)Mode::Simple] *
                                counts[rightSize][iContext][(uint32_t)Mode::Fold];
                    }
                }
            };

            // And
            doOp2(UsedOp::And, Context::And, false);

            // Or
            doOp2(UsedOp::Or, Context::Or, false);

            // Xor
            doOp2(UsedOp::Xor, Context::Xor, false);

            // Plus
            doOp2(UsedOp::Plus, Context::Plus, !(usedOps & UsedOp::Shl1));
        }

        // If0
        if (size >= 4 &&
            !!(usedOps & UsedOp::If0) &&
            (!(usedOpsIn & UsedOp::Bonus) || context == Context::Top))
        {
            for (uint32_t ifSize = (bEasyIf0 ? size - 3 : 1); ifSize + 2 < size; ifSize++)
            {
                if ((usedOpsIn & UsedOp::Bonus) && (ifSize < 5)) continue;
                uint32_t thenElseSize = size - 1 - ifSize;
                for (uint32_t thenSize = 1; thenSize < thenElseSize; thenSize++)
                {
                    if ((usedOpsIn & UsedOp::Bonus) && (thenSize < 5)) continue;
                    uint32_t elseSize = thenElseSize - thenSize;
                    if ((usedOpsIn & UsedOp::Bonus) && (elseSize < 5)) continue;

                    count_t simpleIfCount = counts[ifSize][(uint32_t)Context::If][(uint32_t)Mode::Simple];
                    count_t simpleThenCount = counts[thenSize][(uint32_t)Context::ThenElse][(uint32_t)Mode::Simple];
                    count_t simpleElseCount = counts[elseSize][(uint32_t)Context::ThenElse][(uint32_t)Mode::Simple];

                    count_t tbodyIfCount = counts[ifSize][(uint32_t)Context::If][(uint32_t)Mode::TBody];
                    count_t tbodyThenCount = counts[thenSize][(uint32_t)Context::ThenElse][(uint32_t)Mode::TBody];
                    count_t tbodyElseCount = counts[elseSize][(uint32_t)Context::ThenElse][(uint32_t)Mode::TBody];

                    count_t bodyIfCount = counts[ifSize][(uint32_t)Context::If][(uint32_t)Mode::Body];
                    count_t bodyThenCount = counts[thenSize][(uint32_t)Context::ThenElse][(uint32_t)Mode::Body];
                    count_t bodyElseCount = counts[elseSize][(uint32_t)Context::ThenElse][(uint32_t)Mode::Body];

                    if (thenSize == elseSize)
                    {
                        simpleThenCount--;
                        tbodyThenCount--;
                        bodyThenCount--;
                    }

                    simpleCount +=
                        simpleIfCount * simpleThenCount * simpleElseCount;
                    tbodyCount +=
                        tbodyIfCount * tbodyThenCount * tbodyElseCount;
                    bodyCount +=
                        bodyIfCount * bodyThenCount * bodyElseCount;

                    if (!bFoldFirst)
                        foldCount +=
                            counts[ifSize][(uint32_t)Context::If][(uint32_t)Mode::Fold] *
                            simpleThenCount *
                            counts[elseSize][(uint32_t)Context::ThenElse][(uint32_t)Mode::Simple] +
                            counts[ifSize][(uint32_t)Context::If][(uint32_t)Mode::Simple] *
                            counts[thenSize][(uint32_t)Context::ThenElse][(uint32_t)Mode::Fold] *
                            counts[elseSize][(uint32_t)Context::ThenElse][(uint32_t)Mode::Simple] +
                            counts[ifSize][(uint32_t)Context::If][(uint32_t)Mode::Simple] *
                            counts[thenSize][(uint32_t)Context::ThenElse][(uint32_t)Mode::Simple] *
                            counts[elseSize][(uint32_t)Context::ThenElse][(uint32_t)Mode::Fold];
                }
            }
        }

        // Fold
        if (size >= 3)
        {
            for (uint32_t bytesSize = 1; bytesSize + 1 < size; bytesSize++)
            {
                uint32_t accumSize = size - 1 - bytesSize;

                foldCount +=
                    counts[bytesSize][(uint32_t)Context::FoldBytes][(uint32_t)Mode::Simple] *
                    counts[accumSize][(uint32_t)Context::FoldAccum][(uint32_t)Mode::Simple];
            }
        }

#if 0
        printf("counts[%2"PRIu32"] = %20"PRIu64" %20"PRIu64" %20"PRIu64"\n",
               iCount,
               simpleCount,
               bodyCount,
               count);
#endif
    }
}

count_t GenImpl::getCountFoldBody(uint32_t size)
{
    count_t count = 0;

    if (size == 4)
    {
        count += 20 + 40 + 6;
    }

    return count;
}

count_t GenImpl::getCount()
{
    uint32_t maxSize = sizeIn - 1;
    count_t count = 0;

    for (uint32_t size = (bAllSizes ? 1 : maxSize); size <= maxSize; size++)
    {
        if (usedOpsIn & UsedOp::TFold)
        {
            if (size >= 5)
            {
                count += counts[size - 4][(uint32_t)Context::FoldBody][(uint32_t)Mode::TBody];
            }
        }
        else if (usedOpsIn & UsedOp::Fold)
        {
            if (size >= 5)
            {
                for (uint32_t mainSize = 3; mainSize + 1 < size; mainSize++)
                {
                    uint32_t bodySize = size - 1 - mainSize;

                    Mode bodyMode = (bFoldOnlyYZ ? Mode::TBody : Mode::Body);

                    count_t mainCount = counts[mainSize][(uint32_t)Context::Top][(uint32_t)Mode::Fold];
                    count_t bodyCount =
                        bGenFoldBody ?
                        getCountFoldBody(bodySize) :
                        counts[bodySize][(uint32_t)Context::FoldBody][(uint32_t)bodyMode];

                    count += mainCount * bodyCount;
                }
            }
        }
        else if (usedOpsIn & UsedOp::Bonus)
        {
            count += counts[size][(uint32_t)Context::Top][(uint32_t)Mode::Simple];
        }
        else
        {
            count += counts[size][(uint32_t)Context::Top][(uint32_t)Mode::Simple];
        }
    }

    return count;
}

bool GenImpl::genFoldBody(uint32_t size, Range range, const GenCallback& callback)
{
    if (size != 4) return true;

    pos -= 4;

    const Op op1Arr[5] = { Op::Not, Op::Shl1, Op::Shr1, Op::Shr4, Op::Shr16 };
    const Op op2Arr[4] = { Op::And, Op::Or, Op::Xor, Op::Plus };

    for (uint32_t iOp1 = 0; iOp1 < 5; iOp1++)
    {
        for (uint32_t iOp2 = 0; iOp2 < 4; iOp2++)
        {
            if (range.active(1))
            {
                program.m_ops[pos + 3] = op1Arr[iOp1];
                program.m_ops[pos + 2] = op2Arr[iOp2];
                program.m_ops[pos + 1] = Op::Z;
                program.m_ops[pos + 0] = Op::Y;
                if (!callback(program)) return false;
            }
            range -= 1;
        }
    }

    for (uint32_t iOp2 = 0; iOp2 < 4; iOp2++)
    {
        for (uint32_t iOp1 = 0; iOp1 < 5; iOp1++)
        {
            if (range.active(1))
            {
                program.m_ops[pos + 3] = op2Arr[iOp2];
                program.m_ops[pos + 2] = op1Arr[iOp1];
                program.m_ops[pos + 1] = Op::Z;
                program.m_ops[pos + 0] = Op::Y;
                if (!callback(program)) return false;
            }
            range -= 1;

            if (range.active(1))
            {
                program.m_ops[pos + 3] = op2Arr[iOp2];
                program.m_ops[pos + 2] = op1Arr[iOp1];
                program.m_ops[pos + 1] = Op::Y;
                program.m_ops[pos + 0] = Op::Z;
                if (!callback(program)) return false;
            }
            range -= 1;
        }
    }

    if (range.active(1))
    {
        program.m_ops[pos + 3] = Op::If0;
        program.m_ops[pos + 2] = Op::Z;
        program.m_ops[pos + 1] = Op::Y;
        program.m_ops[pos + 0] = Op::Y;
        if (!callback(program)) return false;
    }
    range -= 1;

    if (range.active(1))
    {
        program.m_ops[pos + 3] = Op::If0;
        program.m_ops[pos + 2] = Op::Y;
        program.m_ops[pos + 1] = Op::Z;
        program.m_ops[pos + 0] = Op::Y;
        if (!callback(program)) return false;
    }
    range -= 1;

    if (range.active(1))
    {
        program.m_ops[pos + 3] = Op::If0;
        program.m_ops[pos + 2] = Op::Z;
        program.m_ops[pos + 1] = Op::Y;
        program.m_ops[pos + 0] = Op::Z;
        if (!callback(program)) return false;
    }
    range -= 1;

    if (range.active(1))
    {
        program.m_ops[pos + 3] = Op::If0;
        program.m_ops[pos + 2] = Op::Y;
        program.m_ops[pos + 1] = Op::Z;
        program.m_ops[pos + 0] = Op::Z;
        if (!callback(program)) return false;
    }
    range -= 1;

    if (range.active(1))
    {
        program.m_ops[pos + 3] = Op::If0;
        program.m_ops[pos + 2] = Op::Y;
        program.m_ops[pos + 1] = Op::Y;
        program.m_ops[pos + 0] = Op::Y;
        if (!callback(program)) return false;
    }
    range -= 1;

    if (range.active(1))
    {
        program.m_ops[pos + 3] = Op::If0;
        program.m_ops[pos + 2] = Op::Z;
        program.m_ops[pos + 1] = Op::Z;
        program.m_ops[pos + 0] = Op::Z;
        if (!callback(program)) return false;
    }
    range -= 1;

    pos += 4;
    return true;
}

bool GenImpl::gen(uint32_t size, Context context, Mode mode, Range range, const GenCallback& callback)
{
    pos--;

    if (size == 1 &&
        (!(usedOpsIn & UsedOp::Bonus) || context != Context::Top))
    {
        // Zero
        if (context != Context::Shl1 &&
            context != Context::Shr1 &&
            context != Context::Shr4 &&
            context != Context::Shr16 &&
            context != Context::Shr1Shl1 &&
            context != Context::Shr4Shl1 &&
            context != Context::Shr16Shl1 &&
            context != Context::And &&
            context != Context::Or &&
            context != Context::Xor &&
            context != Context::AndNot &&
            context != Context::OrNot &&
            context != Context::Plus &&
            context != Context::If &&
            mode != Mode::Fold)
        {
            if (range.active(1))
            {
                program.m_ops[pos] = Op::Zero;
                if (!callback(program)) return false;
            }
            range -= 1;
        }

        // One
        if (context != Context::Shr1 &&
            context != Context::Shr4 &&
            context != Context::Shr16 &&
            context != Context::Shr1Shl1 &&
            context != Context::Shr4Shl1 &&
            context != Context::Shr16Shl1 &&
            context != Context::If &&
            mode != Mode::Fold)
        {
            if (range.active(1))
            {
                program.m_ops[pos] = Op::One;
                if (!callback(program)) return false;
            }
            range -= 1;
        }

        // X
        if (mode == Mode::Simple ||
            mode == Mode::Body)
        {
            if (range.active(1))
            {
                program.m_ops[pos] = Op::X;
                if (!callback(program)) return false;
            }
            range -= 1;
        }

        // Y
        if (mode == Mode::TBody ||
            mode == Mode::Body)
        {
            if (range.active(1))
            {
                program.m_ops[pos] = Op::Y;
                if (!callback(program)) return false;
            }
            range -= 1;
        }

        // Z
        if (mode == Mode::TBody ||
            mode == Mode::Body)
        {
            if (range.active(1))
            {
                program.m_ops[pos] = Op::Z;
                if (!callback(program)) return false;
            }
            range -= 1;
        }
    }

    // Op1
    if (size >= 2 &&
        !(mode == Mode::Fold && bFoldFirst) &&
        (!(usedOpsIn & UsedOp::Bonus) || context != Context::Top))
    {
        auto doOp1 = [&](Op op, uint32_t usedOp, Context context) -> bool {
            if (!(usedOps & usedOp)) return true;

            program.m_ops[pos] = op;
            uint32_t argSize = size - 1;
            count_t argCount = counts[argSize][(uint32_t)context][(uint32_t)mode];
            if (range.active(argCount))
            {
                if (!gen(argSize, context, mode, range, callback)) return false;
            }
            range -= argCount;

            return true;
        };

        // Not
        if (context != Context::Not &&
            context != Context::Xor &&
            context != Context::AndNot &&
            context != Context::OrNot &&
            context != Context::If)
        {
            Context newContext = Context::Not;
            switch (context)
            {
            case Context::And: newContext = Context::AndNot; break;
            case Context::Or: newContext = Context::OrNot; break;
            default: ;
            }
            if (!doOp1(Op::Not, UsedOp::Not, newContext)) return false;
        }

        // Shl1
        {
            Context newContext = Context::Shl1;
            switch (context)
            {
            case Context::Shr1: newContext = Context::Shr1Shl1; break;
            case Context::Shr4: newContext = Context::Shr4Shl1; break;
            case Context::Shr16: newContext = Context::Shr16Shl1; break;
            default: ;
            }
            if (!doOp1(Op::Shl1, UsedOp::Shl1, newContext)) return false;
        }

        // Shr1
        if (context != Context::Shr16 &&
            context != Context::Shr4 &&
            context != Context::Shr1Shl1 &&
            context != Context::Shr4Shl1 &&
            context != Context::Shr16Shl1)
        {
            if (!doOp1(Op::Shr1, UsedOp::Shr1, Context::Shr1)) return false;
        }

        // Shr4
        if (context != Context::Shr16)
        {
            if (!doOp1(Op::Shr4, UsedOp::Shr4, Context::Shr4)) return false;
        }

        // Shr16
        if (!doOp1(Op::Shr16, UsedOp::Shr16, Context::Shr16)) return false;
    }

    // Op2
    if (size >= 3 &&
        !(mode == Mode::Fold && bFoldFirst) &&
        (!(usedOpsIn & UsedOp::Bonus) || context != Context::Top))
    {
        auto doOp2 = [&](Op op, uint32_t usedOp, Context context, bool allowEqual) -> bool {
            program.m_ops[pos] = op;

            if (!(usedOps & usedOp)) return true;

            for (uint32_t leftSize = 1; leftSize + 1 < (bEasyOp2 ? 3 : size); leftSize++)
            {
                uint32_t rightSize = size - 1 - leftSize;

                if (leftSize > rightSize && mode != Mode::Fold)
                {
                    break;
                }
                else if (leftSize == rightSize && mode != Mode::Fold)
                {
                    count_t argCount = counts[leftSize][(uint32_t)context][(uint32_t)mode];
                    if (!allowEqual)
                    {
                        argCount--;
                    }
                    count_t count = (argCount * (argCount + 1)) / 2;
                    if (range.active(count))
                    {
                        count_t leftCount = 1;
                        auto leftCallback = [&](const Program& program) -> bool {
                            if (range.active(leftCount))
                            {
                                Range leftRange(range);
                                if (leftRange.m_high > leftCount)
                                {
                                    leftRange.m_high = leftCount;
                                }
                                if (!gen(leftSize, context, mode, leftRange, callback)) return false;
                            }
                            range -= leftCount;
                            leftCount++;
                            return true;
                        };
                        Range rightRange;
                        if (!allowEqual)
                        {
                            rightRange.m_low = 1;
                        }
                        if (!gen(rightSize, context, mode, rightRange, leftCallback)) return false;
                    }
                    else
                    {
                        range -= count;
                    }
                }
                else
                {
                    Mode leftMode = mode;
                    if (mode == Mode::Fold)
                        leftMode = Mode::Simple;

                    count_t leftCount = counts[leftSize][(uint32_t)context][(uint32_t)leftMode];
                    count_t rightCount = counts[rightSize][(uint32_t)context][(uint32_t)mode];
                    if (range.active(leftCount * rightCount))
                    {
                        auto leftCallback = [&](const Program& program) -> bool {
                            if (range.active(leftCount))
                            {
                                if (!gen(leftSize, context, leftMode, range, callback)) return false;
                            }
                            range -= leftCount;
                            return true;
                        };
                        if (!gen(rightSize, context, mode, Range(), leftCallback)) return false;
                    }
                    else
                    {
                        range -= (leftCount * rightCount);
                    }
                }
            }

            return true;
        };

        // And
        if (!doOp2(Op::And, UsedOp::And, Context::And, false)) return false;

        // Or
        if (!doOp2(Op::Or, UsedOp::Or, Context::Or, false)) return false;

        // Xor
        if (!doOp2(Op::Xor, UsedOp::Xor, Context::Xor, false)) return false;

        // Plus
        if (!doOp2(Op::Plus, UsedOp::Plus, Context::Plus, !(usedOps & UsedOp::Shl1))) return false;
    }

    // If0
    if (size >= 4 &&
        !(mode == Mode::Fold && bFoldFirst) &&
        !!(usedOps & UsedOp::If0) &&
        (!(usedOpsIn & UsedOp::Bonus) || context == Context::Top))
    {
        program.m_ops[pos] = Op::If0;

        for (uint32_t ifSize = (bEasyIf0 ? size - 3 : 1); ifSize + 2 < size; ifSize++)
        {
            if ((usedOpsIn & UsedOp::Bonus) && (ifSize < 5)) continue;
            uint32_t thenElseSize = size - 1 - ifSize;
            for (uint32_t thenSize = 1; thenSize < thenElseSize; thenSize++)
            {
                if ((usedOpsIn & UsedOp::Bonus) && (thenSize < 5)) continue;
                uint32_t elseSize = thenElseSize - thenSize;
                if ((usedOpsIn & UsedOp::Bonus) && (elseSize < 5)) continue;

                Mode thenElseMode = mode;
                if (mode == Mode::Fold)
                {
                    thenElseMode = Mode::Simple;
                }

                if (thenSize == elseSize)
                {
                    count_t ifCount = counts[ifSize][(uint32_t)Context::If][(uint32_t)mode];
                    count_t thenCount = counts[thenSize][(uint32_t)Context::ThenElse][(uint32_t)thenElseMode] - 1;
                    count_t elseCount = counts[elseSize][(uint32_t)Context::ThenElse][(uint32_t)thenElseMode];
                    count_t count = ifCount * thenCount * elseCount;
                    if (range.active(count))
                    {
                        count_t iElse = 0;
                        auto ifThenCallback = [&](const Program& program) -> bool {
                            if (range.active(ifCount * thenCount))
                            {
                                auto ifCallback = [&](const Program& program) -> bool {
                                    if (range.active(ifCount))
                                    {
                                        if (!gen(ifSize, Context::If, mode, range, callback)) return false;
                                    }
                                    range -= ifCount;
                                    return true;
                                };
                                Range thenRange1;
                                thenRange1.m_high = iElse;
                                if (!gen(thenSize, Context::ThenElse, thenElseMode, thenRange1, ifCallback)) return false;
                                Range thenRange2;
                                thenRange2.m_low = iElse + 1;
                                if (!gen(thenSize, Context::ThenElse, thenElseMode, thenRange2, ifCallback)) return false;
                            }
                            else
                            {
                                range -= (ifCount * thenCount);
                            }
                            iElse++;
                            return true;
                        };
                        if (!gen(elseSize, Context::ThenElse, thenElseMode, Range(), ifThenCallback)) return false;
                    }
                    else
                    {
                        range -= (count);
                    }
                }
                else
                {
                    count_t ifCount = counts[ifSize][(uint32_t)Context::If][(uint32_t)mode];
                    count_t thenCount = counts[thenSize][(uint32_t)Context::ThenElse][(uint32_t)thenElseMode];
                    count_t elseCount = counts[elseSize][(uint32_t)Context::ThenElse][(uint32_t)thenElseMode];
                    if (range.active(ifCount * thenCount * elseCount))
                    {
                        auto ifThenCallback = [&](const Program& program) -> bool {
                            if (range.active(ifCount * thenCount))
                            {
                                auto ifCallback = [&](const Program& program) -> bool {
                                    if (range.active(ifCount))
                                    {
                                        if (!gen(ifSize, Context::If, mode, range, callback)) return false;
                                    }
                                    range -= ifCount;
                                    return true;
                                };
                                if (!gen(thenSize, Context::ThenElse, thenElseMode, Range(), ifCallback)) return false;
                            }
                            else
                            {
                                range -= (ifCount * thenCount);
                            }
                            return true;
                        };
                        if (!gen(elseSize, Context::ThenElse, thenElseMode, Range(), ifThenCallback)) return false;
                    }
                    else
                    {
                        range -= (ifCount * thenCount * elseCount);
                    }
                }

                if (mode == Mode::Fold)
                {
                    for (uint32_t iPass = 0; iPass < 2; iPass++)
                    {
                        Mode thenMode = (iPass == 0) ? Mode::Fold : Mode::Simple;
                        Mode elseMode = (iPass == 0) ? Mode::Simple : Mode::Fold;

                        count_t ifCount = counts[ifSize][(uint32_t)Context::If][(uint32_t)Mode::Simple];
                        count_t thenCount = counts[thenSize][(uint32_t)Context::ThenElse][(uint32_t)thenMode];
                        count_t elseCount = counts[elseSize][(uint32_t)Context::ThenElse][(uint32_t)elseMode];
                        if (range.active(ifCount * thenCount * elseCount))
                        {
                            auto ifThenCallback = [&](const Program& program) -> bool {
                                if (range.active(ifCount * thenCount))
                                {
                                    auto ifCallback = [&](const Program& program) -> bool {
                                        if (range.active(ifCount))
                                        {
                                            if (!gen(ifSize, Context::If, Mode::Simple, range, callback)) return false;
                                        }
                                        range -= ifCount;
                                        return true;
                                    };
                                    if (!gen(thenSize, Context::ThenElse, thenMode, Range(), ifCallback)) return false;
                                }
                                else
                                {
                                    range -= (ifCount * thenCount);
                                }
                                return true;
                            };
                            if (!gen(elseSize, Context::ThenElse, elseMode, Range(), ifThenCallback)) return false;
                        }
                        else
                        {
                            range -= (ifCount * thenCount * elseCount);
                        }
                    }
                }
            }
        }
    }

    // Fold
    if (size >= 3 &&
        mode == Mode::Fold)
    {
        program.m_ops[pos] = Op::Fold;

        for (uint32_t bytesSize = 1; bytesSize + 1 < size; bytesSize++)
        {
            uint32_t accumSize = size - 1 - bytesSize;

            count_t bytesCount = counts[bytesSize][(uint32_t)Context::FoldBytes][(uint32_t)Mode::Simple];
            count_t accumCount = counts[accumSize][(uint32_t)Context::FoldAccum][(uint32_t)Mode::Simple];
            if (range.active(bytesCount * accumCount))
            {
                auto bytesCallback = [&](const Program& program) -> bool {
                    if (range.active(bytesCount))
                    {
                        if (!gen(bytesSize, Context::FoldBytes, Mode::Simple, range, callback)) return false;
                    }
                    range -= bytesCount;
                    return true;
                };
                if (!gen(accumSize, Context::FoldAccum, Mode::Simple, Range(), bytesCallback)) return false;
            }
            else
            {
                range -= (bytesCount * accumCount);
            }
        }
    }

    pos++;
    return true;
}

bool GenImpl::generate(const GenCallback& callback)
{
    uint32_t maxSize = sizeIn - 1;

    Range range;

    for (uint32_t size = (bAllSizes ? 1 : maxSize); size <= maxSize; size++)
    {
        if (usedOpsIn & UsedOp::TFold)
        {
            if (size >= 5)
            {
                pos = size - 1;
                program.m_ops[0] = Op::X;
                program.m_ops[1] = Op::Zero;
                program.m_ops[2] = Op::Fold;
                program.m_opEnd = 3;
                program.m_foldOpEnd = pos;

                count_t count = counts[size - 4][(uint32_t)Context::FoldBody][(uint32_t)Mode::TBody];
                if (range.active(count))
                {
                    if (!gen(size - 4, Context::FoldBody, Mode::TBody, range, callback)) return false;
                }
                range -= count;
            }
        }
        else if (usedOpsIn & UsedOp::Fold)
        {
            if (size >= 5)
            {
                for (uint32_t mainSize = 3; mainSize + 1 < size; mainSize++)
                {
                    uint32_t bodySize = size - 1 - mainSize;

                    pos = size - 1;
                    program.m_opEnd = mainSize;
                    program.m_foldOpEnd = size - 1;

                    Mode bodyMode = (bFoldOnlyYZ ? Mode::TBody : Mode::Body);

                    count_t mainCount = counts[mainSize][(uint32_t)Context::Top][(uint32_t)Mode::Fold];
                    count_t bodyCount =
                        bGenFoldBody ?
                        getCountFoldBody(bodySize) :
                        counts[bodySize][(uint32_t)Context::FoldBody][(uint32_t)bodyMode];
                    if (range.active(mainCount * bodyCount))
                    {
                        auto mainCallback = [&](const Program& program) -> bool {
                            if (range.active(mainCount))
                            {
                                if (!gen(mainSize, Context::Top, Mode::Fold, range, callback)) return false;
                            }
                            range -= mainCount;
                            return true;
                        };
                        if (bGenFoldBody)
                        {
                            if (!genFoldBody(bodySize, Range(), mainCallback)) return false;
                        }
                        else
                        {
                            if (!gen(bodySize, Context::FoldBody, bodyMode, Range(), mainCallback)) return false;
                        }
                    }
                    else
                    {
                        range -= (mainCount * bodyCount);
                    }
                }
            }
        }
        else if (usedOpsIn & UsedOp::Bonus)
        {
            pos = size;
            program.m_opEnd = pos;
            program.m_foldOpEnd = pos;

            count_t count = counts[size][(uint32_t)Context::Top][(uint32_t)Mode::Simple];
            if (range.active(count))
            {
                if (!gen(size, Context::Top, Mode::Simple, range, callback)) return false;
            }
            range -= count;
        }
        else
        {
            pos = size;
            program.m_opEnd = pos;
            program.m_foldOpEnd = pos;

            count_t count = counts[size][(uint32_t)Context::Top][(uint32_t)Mode::Simple];
            if (range.active(count))
            {
                if (!gen(size, Context::Top, Mode::Simple, range, callback)) return false;
            }
            range -= count;
        }
    }

    return true;
}

Generator::Generator() : m_pImpl(new GenImpl)
{
}

Generator::~Generator()
{
}

void Generator::init(uint32_t size, uint32_t usedOps)
{
    m_pImpl->init(size, usedOps);
}

count_t Generator::getCount()
{
    return m_pImpl->getCount();
}

void Generator::generate(const GenCallback& callback)
{
    m_pImpl->generate(callback);
}
