#ifndef PROTOCOL_H
#define PROTOCOL_H

#include <curl/curl.h>
#include <string>
#include <vector>
#include <stdint.h>

class Problem;

enum class TrainOps : uint32_t
{
    Unspecified,
    None,
    TFold,
    Fold
};

enum class GuessStatus : uint32_t
{
    Win,
    Mismatch,
    Error
};

enum class GuessLightning : uint32_t
{
    Unspecified,
    False,
    True
};

namespace Protocol
{
    void init();

    bool getProblems(std::vector<Problem>* pProblems,
                     std::string* pMsg = nullptr);
    bool getProblems(const std::string& filename,
                     std::string* pMsg = nullptr);
    bool loadProblems(const std::string& filename,
                      std::vector<Problem>* pProblems,
                      std::string* pMsg = nullptr);

    bool evalId(const std::string& id,
                const std::vector<uint64_t>& args,
                std::vector<uint64_t>* pOutputs,
                std::string* pMsg = nullptr);
    bool evalProg(const std::string& prog,
                  const std::vector<uint64_t>& args,
                  std::vector<uint64_t>* pOutputs,
                  std::string* pMsg = nullptr);

    bool train(uint32_t size,
               TrainOps ops,
               Problem* pProblem,
               std::string* pChallenge,
               std::string* pMsg = nullptr);

    bool guess(const std::string& id,
               const std::string& strProgram,
               GuessStatus* pStatus,
               uint64_t* pInput,
               uint64_t* pChallengeOutput,
               uint64_t* pGuessOutput,
               std::string* pError,
               GuessLightning* pLightning,
               std::string* pMsg = nullptr);
}

#endif
