#include "Problem.hpp"
#include "Program.hpp"
#include "Protocol.hpp"
#include "Generator.hpp"
#include "Solver.hpp"
#include <stdio.h>
#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <vector>
#include <string>
#include <algorithm>
#include <time.h>
#include <unistd.h>
#include <string.h>

using std::vector;
using std::string;

class ProblemComp
{
public:
    bool operator() (const Problem& problem1, const Problem& problem2)
    {
        if (problem1.m_size != problem2.m_size)
        {
            return problem1.m_size < problem2.m_size;
        }

        if ((problem1.m_usedOps & UsedOp::Bonus) !=
            (problem2.m_usedOps & UsedOp::Bonus))
        {
            return
                (problem1.m_usedOps & UsedOp::Bonus) <
                (problem2.m_usedOps & UsedOp::Bonus);
        }

        if ((problem1.m_usedOps & UsedOp::Fold) !=
            (problem2.m_usedOps & UsedOp::Fold))
        {
            return
                (problem1.m_usedOps & UsedOp::Fold) <
                (problem2.m_usedOps & UsedOp::Fold);
        }

        if ((problem1.m_usedOps & UsedOp::TFold) !=
            (problem2.m_usedOps & UsedOp::TFold))
        {
            return
                (problem1.m_usedOps & UsedOp::TFold) <
                (problem2.m_usedOps & UsedOp::TFold);
        }

        uint32_t usedOpsCount1 = 0;
        if (problem1.m_usedOps & UsedOp::Not) usedOpsCount1++;
        if (problem1.m_usedOps & UsedOp::Shl1) usedOpsCount1++;
        if (problem1.m_usedOps & UsedOp::Shr1) usedOpsCount1++;
        if (problem1.m_usedOps & UsedOp::Shr4) usedOpsCount1++;
        if (problem1.m_usedOps & UsedOp::Shr16) usedOpsCount1++;
        if (problem1.m_usedOps & UsedOp::And) usedOpsCount1++;
        if (problem1.m_usedOps & UsedOp::Or) usedOpsCount1++;
        if (problem1.m_usedOps & UsedOp::Xor) usedOpsCount1++;
        if (problem1.m_usedOps & UsedOp::Plus) usedOpsCount1++;
        if (problem1.m_usedOps & UsedOp::If0) usedOpsCount1++;

        uint32_t usedOpsCount2 = 0;
        if (problem2.m_usedOps & UsedOp::Not) usedOpsCount2++;
        if (problem2.m_usedOps & UsedOp::Shl1) usedOpsCount2++;
        if (problem2.m_usedOps & UsedOp::Shr1) usedOpsCount2++;
        if (problem2.m_usedOps & UsedOp::Shr4) usedOpsCount2++;
        if (problem2.m_usedOps & UsedOp::Shr16) usedOpsCount2++;
        if (problem2.m_usedOps & UsedOp::And) usedOpsCount2++;
        if (problem2.m_usedOps & UsedOp::Or) usedOpsCount2++;
        if (problem2.m_usedOps & UsedOp::Xor) usedOpsCount2++;
        if (problem2.m_usedOps & UsedOp::Plus) usedOpsCount2++;
        if (problem2.m_usedOps & UsedOp::If0) usedOpsCount2++;

        if (usedOpsCount1 != usedOpsCount2)
        {
            return usedOpsCount1 < usedOpsCount2;
        }

        return problem1.m_id.compare(problem2.m_id) < 0;
    }
};

void usage()
{
    printf("Usage: do [list | dump | time | solve | index] [all | <id> | <size 3..42>] [none | tfold | fold | bonus]\n");
}

int main(int argc, char *argv[])
{
    Protocol::init();

    if (argc < 3 || argc > 4)
    {
        usage();
        return 1;
    }

    bool bList = false;
    bool bDump = false;
    bool bTime = false;
    bool bSolve = false;
    bool bIndex = false;
    if (strcmp(argv[1], "list") == 0)
    {
        bList = true;
    }
    else if (strcmp(argv[1], "dump") == 0)
    {
        bDump = true;
    }
    else if (strcmp(argv[1], "time") == 0)
    {
        bTime = true;
    }
    else if (strcmp(argv[1], "solve") == 0)
    {
        bSolve = true;
    }
    else if (strcmp(argv[1], "index") == 0)
    {
        bIndex = true;
    }
    else
    {
        usage();
        return 1;
    }
    (void)bList;

    bool bAll = false;
    bool bId = false;
    string id;
    bool bSize = false;
    uint32_t size = 0;
    if (strcmp(argv[2], "all") == 0)
    {
        bAll = true;
    }
    else
    {
        char* endptr = nullptr;
        size = strtol(argv[2], &endptr, 10);
        if (*argv[2] && !*endptr && size >= 0 && size <= 1000)
        {
            if (size < 3 || size > 42)
            {
                usage();
                return 1;
            }
            bSize = true;
        }
        else
        {
            bId = true;
            id = argv[2];
        }
    }
    (void)bAll;

    bool bType = false;
    bool bNone = false;
    bool bTFold = false;
    bool bFold = false;
    bool bBonus = false;
    if (argc > 3)
    {
        bType = true;
        if (strcmp(argv[3], "none") == 0)
        {
            bNone = true;
        }
        else if (strcmp(argv[3], "tfold") == 0)
        {
            bTFold = true;
        }
        else if (strcmp(argv[3], "fold") == 0)
        {
            bFold = true;
        }
        else if (strcmp(argv[3], "bonus") == 0)
        {
            bBonus = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (bSolve)
    {
        if (!bId && !bSize)
        {
            printf("ID or size required to solve\n");
            return 1;
        }

        if (!bType)
        {
            printf("Type required to solve\n");
            return 1;
        }
    }

    string msg;
    vector<Problem> problems;
    if (!Protocol::loadProblems("../data/myproblems.json", &problems, &msg))
    {
        printf("Error loading problems: %s\n", msg.c_str());
        return 1;
    }

    uint32_t problemCount = (uint32_t)problems.size();
    //printf("problemCount = %"PRIu32"\n", problemCount);

    std::sort(problems.begin(), problems.end(), ProblemComp());

    for (uint32_t iProblem = 0; iProblem < problemCount; iProblem++)
    {
        const Problem& problem = problems[iProblem];

        if (bId)
        {
            if (id.compare(problem.m_id) != 0)
            {
                continue;
            }
        }

        if (bSize)
        {
            if (problem.m_size != size)
            {
                continue;
            }
        }

        if (bNone)
        {
            if (problem.m_usedOps & (UsedOp::TFold | UsedOp::Fold | UsedOp::Bonus))
            {
                continue;
            }
        }
        if (bTFold)
        {
            if (!(problem.m_usedOps & UsedOp::TFold))
            {
                continue;
            }
        }
        if (bFold)
        {
            if (!(problem.m_usedOps & UsedOp::Fold))
            {
                continue;
            }
        }
        if (bBonus)
        {
            if (!(problem.m_usedOps & UsedOp::Bonus))
            {
                continue;
            }
        }

        const char* strOps = "none";
        if (problem.m_usedOps & UsedOp::Fold) strOps = "fold";
        if (problem.m_usedOps & UsedOp::TFold) strOps = "tfold";
        if (problem.m_usedOps & UsedOp::Bonus) strOps = "bonus";

        const char* strSolved = "?";
        if (problem.m_solved == ProblemSolved::True) strSolved = "$";
        if (problem.m_solved == ProblemSolved::False) strSolved = "_";

        printf("%s, %s, %2"PRIu32", %s, 0x%04"PRIx32": ",
               problem.m_id.c_str(),
               strSolved,
               problem.m_size,
               strOps,
               problem.m_usedOps &
               ~(UsedOp::Fold | UsedOp::TFold | UsedOp::Bonus));
        fflush(stdout);

        if (bDump)
        {
            printf("\n");
        }

        FILE* indexFile = nullptr;
        string strIndexFile;
        string strSortFile;
        if (bIndex)
        {
            strIndexFile = "../../../index/";
            strIndexFile += problem.m_id;
            strSortFile = strIndexFile;
            strIndexFile += ".dat";
            strSortFile += ".idx";
            indexFile = fopen(strIndexFile.c_str(), "wb");
            if (!indexFile)
            {
                printf("\n");
                printf("Error opening index file\n");
                return 1;
            }
        }

        uint64_t count = 0;
        auto callback = [&](const Program& program) -> bool {
            //uint64_t iProgram = count;
            count++;
            if (bDump)
            {
                printf("  %s\n", program.toString().c_str());
            }

            if (bIndex)
            {
                uint64_t data[16];
                data[ 0] = program.eval(0x428a2f98d728ae22);
                data[ 1] = program.eval(0x0000000000000000);
                data[ 2] = program.eval(0x7137449123ef65cd);
                data[ 3] = program.eval(0xffffffffffffffff);
                data[ 4] = program.eval(0xb5c0fbcfec4d3b2f);
                data[ 5] = program.eval(0x0000000000000001);
                data[ 6] = program.eval(0xe9b5dba58189dbbc);
                data[ 7] = program.eval(0xfffffffffffffffe);
                data[ 8] = program.eval(0x0000000000000010);
                data[ 9] = program.eval(0x0000000000020000);
                data[10] = program.eval(0x0000000400000000);
                data[11] = program.eval(0x8000000000000000);

                program.pack(&data[12]);

                if (fwrite(data, sizeof(char), 128, indexFile) != 128)
                {
                    printf("\n");
                    printf("Error writing index file\n");
                    return false;
                }
            }
            else
            {
                const uint32_t valueCount = 8;
                for (uint32_t iValue = 0; iValue < valueCount; iValue++)
                {
                    program.eval(0x0123456789abcdef);
                }
            }

            return true;
        };

        Generator generator;
        generator.init(problem.m_size, problem.m_usedOps);

        if (bTime || bDump || bIndex)
        {
            time_t startTime = time(nullptr);
            generator.generate(callback);
            time_t endTime = time(nullptr);
            uint32_t elapsed = (uint32_t)(endTime - startTime);

            if (!bDump)
            {
                printf("%1"PRIu32":%02"PRIu32", %"PRIu64", ",
                       elapsed / 60, elapsed % 60, count);
            }
        }

        count_t newCount = generator.getCount();

        if (!bDump)
        {
            printf("%"PRIu64"", (uint64_t)newCount);
        }

        if (!bDump)
        {
            printf("\n");
        }

        if (bIndex)
        {
            fclose(indexFile);

            //printf("Sorrting...\n");
        }

        if (bSolve)
        {
#if 1
            if (!solve(problem))
            {
                printf("Solve failed!\n");
                return 1;
            }
            sleep(6);
#else
            printf("Solving disabled\n");
#endif
        }
    }

    return 0;
}
