#include "Problem.hpp"
#include "Protocol.hpp"
#include "Solver.hpp"
#include <stdio.h>
#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <vector>
#include <string>
#include <string.h>

using std::vector;
using std::string;

void usage()
{
    printf("Usage: solve <id> [index]\n");
}

int main(int argc, char *argv[])
{
    Protocol::init();

    if (argc < 2 || argc > 3)
    {
        usage();
        return 1;
    }

    string id(argv[1]);

    bool bIndex = false;
    if (argc > 2)
    {
        if (strcmp(argv[2], "index") == 0)
        {
            bIndex = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    string msg;
    vector<Problem> problems;
    if (!Protocol::loadProblems("../data/myproblems.json", &problems, &msg))
    {
        printf("Error loading problems: %s\n", msg.c_str());
        return 1;
    }

    uint32_t problemCount = (uint32_t)problems.size();

    uint32_t iFoundProblem = 0;
    bool found = false;
    for (uint32_t iProblem = 0; iProblem < problemCount; iProblem++)
    {
        const Problem& problem = problems[iProblem];
        if (id.compare(problem.m_id) == 0)
        {
            iFoundProblem = iProblem;
            found = true;
            break;
        }
    }

    if (!found)
    {
        printf("Problem id not found\n");
        return 1;
    }

    const Problem& problem = problems[iFoundProblem];

    const char* strOps = "none";
    if (problem.m_usedOps & UsedOp::Fold) strOps = "fold";
    if (problem.m_usedOps & UsedOp::TFold) strOps = "tfold";
    if (problem.m_usedOps & UsedOp::Bonus) strOps = "bonus";

    const char* strSolved = "?";
    if (problem.m_solved == ProblemSolved::True) strSolved = "$";
    if (problem.m_solved == ProblemSolved::False) strSolved = "_";

    printf("%s, %s, %2"PRIu32", %s, 0x%04"PRIx32"\n",
           problem.m_id.c_str(),
           strSolved,
           problem.m_size,
           strOps,
           problem.m_usedOps &
           ~(UsedOp::Fold | UsedOp::TFold | UsedOp::Bonus));

    bool solveResult = false;
    if (bIndex)
    {
        string filename("../../../index/");
        filename += problem.m_id;
        filename += ".idx";
        solveResult = indexSolve(problem, filename);
    }
    else
    {
        solveResult = solve(problem);
    }

    if (!solveResult)
    {
        printf("Solve failed!\n");
        return 1;
    }

    return 0;
}
