#include "Problem.hpp"
#include "Protocol.hpp"
#include "Solver.hpp"
#include <stdio.h>
#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <vector>
#include <string>
#include <string.h>

using std::vector;
using std::string;

void usage()
{
    printf("Usage: train <size 3..137> [none | tfold | fold]\n");
}

int main(int argc, char *argv[])
{
    Protocol::init();

    if (argc < 2 || argc > 3)
    {
        usage();
        return 1;
    }

    uint32_t size = (uint32_t)atol(argv[1]);
    if (size < 3 || size > 137)
    {
        usage();
        return 1;
    }

    TrainOps trainOps = TrainOps::Unspecified;
    if (argc >= 3)
    {
        if (strcmp("none", argv[2]) == 0)
        {
            trainOps = TrainOps::None;
        }
        else if (strcmp("tfold", argv[2]) == 0)
        {
            trainOps = TrainOps::TFold;
        }
        else if (strcmp("fold", argv[2]) == 0)
        {
            trainOps = TrainOps::Fold;
        }
        else
        {
            usage();
            return 1;
        }
    }

    string msg;

    Problem problem;
    string challenge;
    if (!Protocol::train(size, trainOps, &problem, &challenge, &msg))
    {
        printf("train failed: %s\n", msg.c_str());
        return 1;
    }

    printf("Challenge: %s\n", challenge.c_str());
    printf("%s, %2"PRIu32", 0x%04"PRIx32"\n",
           problem.m_id.c_str(),
           problem.m_size,
           problem.m_usedOps);

    if (!solve(problem))
    {
        return 1;
    }


    return 0;
}
